"""Provides DAG functionality used by chores and k8s job sched
"""
from collections import defaultdict
from copy import deepcopy
from itertools import chain
from typing import DefaultDict, Set, List, NewType, Union
from cachetools import cached
from cachetools.keys import hashkey


Node = NewType('Node', Union[str, int])
Graph = NewType('Graph', DefaultDict[Node, Set[Node]])


def topo_sort(froms: Graph) -> List[Node]:
    """Sort a graph where there is an edge from each key to their values.

    :param froms: A Graph where there exists a directed edge from k to each
        v in Graph[k], k -> v

    :returns: A list of nodes
    """
    tos = rev_graph(froms)
    return topo_sort_kahn(deepcopy(froms), tos)


def topo_sort_parallel(froms: Graph) -> List[Set[Node]]:
    """Sort a graph where there is an edge from key to values.

    :param froms: A Graph where there exists a directed edge from k to each
        v in Graph[k], k -> v

    :returns: A list of sets of nodes where each node in a set can be
        executed in parallel.
    """
    tos = rev_graph(froms)
    return _topo_sort_parallel(tos)


def topo_sort_kahn(froms: Graph, tos: Graph) -> List[Node]:
    """Do a topological sort with Kahn's algo.

    :param froms: A Graph where there exists a directed edge from k to each
        v in Graph[k], k -> v
    :param tos: A Graph where there exists a directed edge from each value
        v in Graph[k] to k, v -> k
    :raises ValueError: If the graph has a cycle in it.
    :returns: The list of nodes in topological order.
    """
    sort = []
    no_ins = set(nd for nd in chain(tos, froms) if not tos[nd])
    while no_ins:
        nd = no_ins.pop()
        sort.append(nd)
        for nd2 in froms[nd]:
            tos[nd2].discard(nd)
            if not tos[nd2]:
                no_ins.add(nd2)
        froms[nd] = set()
    for ed in froms.values():
        if ed:
            raise ValueError('Graph has a cycle')
    return sort


class CycleError(ValueError):
    """An error class to use if the DAG has a cycle in it."""


def _topo_sort_parallel(tos: Graph) -> List[Set[Node]]:
    """Do a topological sort that returns sets of parallel possible jobs.

    :param tos: A Graph where there exists a directed edge from each value
        v in Graph[k] to k, v -> k
    :raises CycleError: If the graph has a cycle in it.
    :returns: A list of sets where each set can be run in parallel.
    """
    sort = []
    while True:
        no_ins = set(n for n, pre in tos.items() if not pre)
        if not no_ins:
            break
        sort.append(no_ins)
        tos = {n: pre - no_ins for n, pre in tos.items() if n not in no_ins}
    if tos:
        raise CycleError('Graph has a cycle')
    return sort


def find_children(froms: Graph) -> DefaultDict[Node, Set[Node]]:
    """Find all the children of all nodes.

    :param froms: A graph where there is an edge from each k to each value v
        in graph[k], k -> v

    :returns: A map for nodes to children.
    """
    return defaultdict(set, {n: _find_children(froms, n) for n in froms})


# The graph is unhashable so we need to stringify it so we can cache it.
@cached(cache={}, key=lambda froms, root: (hashkey(str(froms)), hashkey(root)))
def _find_children(froms: Graph, root: Node) -> Set[Node]:
    """Find all the children of some node.

    :param froms: A graph where there is an edge from each k to each value v
        in graph[k], k -> v
    :param root: The node to collect the children from.

    :returns: The children of `root`
    """
    descendents = set(froms[root])
    for child in froms[root]:
        descendents.update(_find_children(froms, child))
    return descendents


def rev_graph(graph: Graph) -> Graph:
    """Convert a graph from incoming to outgoing.
    :param graph: A graph to reverse
    :returns: A reversed graph
    """
    r_graph = defaultdict(set)
    for nd, eds in graph.items():
        for ed in eds:
            r_graph[ed].add(nd)
        # Make sure this node is in the graph even if it has no edges
        r_graph[nd]  # pylint: disable=W0104
    return r_graph


def dot_graph(froms: Graph) -> str:
    """Convert a graph into a dot string.

    :param froms: A Graph where there exists a directed edge from k to each
        v in Graph[k], k -> v
    :returns: A dot string representing the graph
    """
    lines = ['digraph {']
    for nd, eds in froms.items():
        nd = str(nd).replace('-', '_')
        if not eds:
            lines.append(f"  {nd};")
        for ed in eds:
            ed = str(ed)
            lines.append(f"  {nd} -> {ed.replace('-', '_')};")
    lines.append("}")
    return "\n".join(lines)


def write_graph(froms: Graph, file_name: str) -> None:
    """Write a graph to file in dot format.

    :param froms: A Graph where there exists a directed edge from k to each
        v in Graph[k], k -> v
    :param file_name: A file name to write to
    """
    with open(file_name, 'w') as write_file:
        write_file.write(dot_graph(froms))



"""Module to help decide what jobs are worthy of accepting

This module provides an interface for `ExportPolicy` and
various implementations that use XPCTL

It also includes a command-line driver to run the export policies
"""
import argparse
import logging
import json
from operator import gt, ge, itemgetter
from typing import List, Dict, Tuple, Callable, Type, Optional, Union

from baseline.utils import exporter, get_metric_cmp, optional_params, read_config_stream
from mead.utils import convert_path, get_dataset_from_key
from odin.store import Store, create_store_backend
from xpclient.utils import xpctl_client
from xpclient.api import XpctlApi
from xpclient.rest import ApiException


__all__ = []
export = exporter(__all__)
EXPORT_POLICY_REGISTRY = {}

LOGGER = logging.getLogger('odin')


@export
class ExportPolicy:
    """See several metrics and decide which (if any should be exported)."""

    def __init__(self) -> None:
        """Constructor"""

    def select(self, job_ids: List[str]) -> Dict:
        """Evaluate a trained model based on this policy.

        :param job_ids: List of jobs to evaluate results from

        :returns: A dictionary of results
        """


@export
@optional_params
def register_export_policy(cls: Type[ExportPolicy], name: str = None) -> Type[ExportPolicy]:
    """Register an export policy

    :param cls: The class name
    :param name: The registry name
    :raises Exception: If name is already registered.
    :return: The class
    """
    name = name if name is not None else cls.__name__
    if name in EXPORT_POLICY_REGISTRY:
        raise Exception(
            'Error: attempt to re-define previously '
            f'registered ExportPolicy {name} (old: {EXPORT_POLICY_REGISTRY[name]}, new: {cls}) in registry'
        )
    EXPORT_POLICY_REGISTRY[name] = cls
    return cls


@export
def create_export_policy(name: str, config: Dict) -> ExportPolicy:
    """Create an export policy from the registry

    :param name: A name identifier
    :param config: The configuration to pass
    :return: A constructed instance
    """
    return EXPORT_POLICY_REGISTRY[name](**config)


@export
class ExperimentRepoExportPolicy(ExportPolicy):
    """Select the best model scoring model of the ones seen."""

    DATASET = 'dataset'
    LABEL = 'label'

    def __init__(self, api: XpctlApi, task: str, dataset: Union[str, None], metric: str, **kwargs) -> None:
        """Constructor for experiment repo-based exporting

        :param api: An XPCTL client api object
        :param task: A string task name
        :param dataset: A string key for dataset
        :param metric: The metric to evaluate
        :param kwargs:
        """
        super().__init__()
        self.api = api
        if task is None:
            raise ValueError("`task` parameter must not be None")
        self.task = task
        self.dataset: str = self._find_dataset(dataset) if dataset is not None else None
        self.metric: str = metric
        self.best_value: float = None
        self.best_label: float = None

    def _create_result_dict(self, job_id):
        return {
            "value": self.best_value,
            "selected": job_id,
            "metric": self.metric,
            "task": self.task,
            "dataset": self.dataset,
        }

    def _find_dataset(self, dataset):
        datasets = self.api.task_summary(self.task)
        datasets = {k: k for k in datasets.summary}
        return get_dataset_from_key(dataset, datasets)


@export
@register_export_policy('best-of-batch')
class BestOfBatchExportPolicy(ExperimentRepoExportPolicy):
    """Select a model that is the best of the batch"""

    def __init__(self, api: XpctlApi, task: str, dataset: str, metric: str, user_cmp: str = None, **kwargs) -> None:
        """Construct a best model in batch

        :param api: An XPCTL client api object
        :param task: A string name for the task
        :param dataset: A string key for the dataset
        :param metric: The metric to evaluate
        :param user_cmp: A comparator function
        :param kwargs:
        """
        super().__init__(api, task, dataset=None, metric=metric)
        self.cmp: Callable[[float, float], bool] = None
        self.cmp, self.best_value = get_metric_cmp(self.metric, user_cmp)

    def _get_by_label(self, job_id: str) -> float:
        """
        Get the value for a metric for an experiment. The experiment is found by the label.

        :param job_id: id for a job
        :raises RuntimeError: ApiException indicates we failed to connect to xpctl server.
        :return: the value for a metric for an experiment
        """
        try:
            exps = self.api.list_experiments_by_prop(self.task, label=job_id)
            return [x.value for x in exps[0].test_events if x.metric == self.metric][0]
            # assuming labels are unique here
        except ApiException as exception:
            raise RuntimeError(json.loads(exception.body)['detail'])

    def select(self, job_ids: List[str]) -> Dict:
        """Select the best job in the batch
        :param job_ids: The list of job IDs
        :return: The result dictionary
        """
        best_label = None
        for job_id in job_ids:
            value = self._get_by_label(job_id)
            if self.cmp(value, self.best_value):
                self.best_value = value
                best_label = job_id

        return self._create_result_dict(best_label)


@export
@register_export_policy('best-of-all')
class BestExportPolicy(ExperimentRepoExportPolicy):
    """Select a model that is the best on this dataset"""

    def __init__(self, api: XpctlApi, task: str, dataset: str, metric: str, user_cmp: str = None, **kwargs) -> None:
        """

        :param api: An XPCTL client api object
        :param task: A string name for the task
        :param dataset: A string key for the dataset
        :param metric: The dataset to evaluate
        :param user_cmp: A comparator function
        :param kwargs:
        """
        assert dataset is not None
        super().__init__(api, task, dataset, metric)
        self.cmp: Callable[[float, float], bool] = None
        self.cmp, self.best_value = get_metric_cmp(self.metric, user_cmp)

    def _get_results(self) -> List[Tuple[str, float]]:
        """
        Return List[(label, value)] for all experiments with this dataset.

        :raises RuntimeError: ApiException indicates we failed to connect to xpctl server.
        :return: Give back a tuple of the label and the score
        """
        try:
            exps = self.api.list_experiments_by_prop(task=self.task, dataset=self.dataset)
            labels_values = []
            for exp in exps:
                labels_values.append((exp.label, [x.value for x in exp.test_events if x.metric == self.metric][0]))
            return labels_values
        except ApiException as exception:
            raise RuntimeError(json.loads(exception.body)['detail'])

    def select(self, job_ids: List[str]) -> Dict:
        """Select the best job in the dataset
        :param job_ids: The list of job IDs
        :return: The result dictionary
        """
        labels_values = self._get_results()
        best_job = None
        for job_id, job_value in labels_values:
            if self.cmp(job_value, self.best_value) and job_id in job_ids:
                self.best_value = job_value
                best_job = job_id
        if best_job is not None:
            return self._create_result_dict(best_job)
        return {}


@export
@register_export_policy('best-of-mead-eval-only')
class BestExportPolicyOutput(ExportPolicy):
    """Select the best model based on mead eval results.

    This is actually more general it selects the best based on
    values in the outputs section of a job db entry.
    """

    def __init__(self, store: Store, metric: str, user_cmp: Optional[str] = None, **kwargs) -> None:
        """
        :param store: The odin job db
        :param metric: The metric used to compare models
        :param user_cmp: A user supplied comparison operator
        :param kwargs: absorb kwargs from other exports
        """
        super().__init__()
        self.store = store
        self.metric = metric
        cmp, self.default_value = get_metric_cmp(self.metric, user_cmp)
        self.reverse = cmp in {gt, ge}
        self.best_value = 0

    def _get_result(self, job_id: str, metric: str) -> float:
        """Get the metric out of the job db entry.

        :param job_id: The job to look up
        :param metric: The metric to look for
        :returns: The metric
        """
        job_entry = self.store.get(job_id)
        return job_entry.get('outputs', {}).get(metric, self.default_value)

    def _create_result_dict(self, job_id: str) -> Dict:
        return {"value": self.best_value, "selected": job_id, "metric": self.metric}

    def select(self, job_ids: List[str]) -> Dict:
        """Select the best job in the dataset
        :param job_ids: The list of job IDs
        :return: The result dictionary
        """
        job = [(j, self._get_result(j, self.metric)) for j in job_ids]
        best_job = sorted(job, key=itemgetter(1), reverse=self.reverse)[0]
        self.best_value = best_job[1]
        return self._create_result_dict(best_job[0])


@export
@register_export_policy('best-of-mead-eval')
class BestExporrPolicyMeadEvalXPCTL(ExperimentRepoExportPolicy):
    """Get the best model sorted by mead eval and then xpctl scores."""

    def __init__(
        self, api: XpctlApi, store: Store, task: str, dataset: str, metric: str, user_cmp: str = None, **kwargs
    ) -> None:
        """Construct a best model in batch

        :param api: An XPCTL client api object
        :param store: The odin jobs db
        :param task: A string name for the task
        :param dataset: A string key for the dataset
        :param metric: The metric to evaluate
        :param user_cmp: A comparator function
        :param kwargs:
        """
        super().__init__(api, task, dataset, metric)
        self.store = store
        self.cmp: Callable[[float, float], bool] = None
        self.cmp, self.default_value = get_metric_cmp(self.metric, user_cmp)
        self.reverse = self.cmp in {gt, ge}

    def _get_xpctl_result(self, job_id: str) -> float:
        """
        Get the value for a metric for an experiment. The experiment is found by the label.

        :param job_id: id for a job
        :raises RuntimeError: ApiException indicates we failed to connect to xpctl server.
        :return: the value for a metric for an experiment
        """
        try:
            exps = self.api.list_experiments_by_prop(self.task, label=job_id)
            return [x.value for x in exps[0].test_events if x.metric == self.metric][0]
            # assuming labels are unique here
        except ApiException as exception:
            raise RuntimeError(json.loads(exception.body)['detail'])

    def _get_results(self, job_id: str, metric: str) -> Tuple[float, float]:
        """Get the mead-eval output for a job and if the output includes
           the label get the xpctl score too.

        :param job_id: The job to look up the results for
        :param metric: The metric to look up.
        :returns: The mead-eval and xpctl scores
        """
        job_entry = self.store.get(job_id)
        outputs = job_entry.get('outputs', {})
        mead_eval = outputs.get(metric, self.default_value)
        label = outputs.get('label')
        xpctl = self._get_xpctl_result(label) if label is not None else self.default_value
        return mead_eval, xpctl

    def select(self, job_ids: List[str]) -> Dict:
        """Select the best job in the dataset
        :param job_ids: The list of job IDs
        :return: The result dictionary
        """
        job = [(j, *self._get_results(j, self.metric)) for j in job_ids]
        # Python has a stable sort so if we sort by xpctl first
        job = sorted(job, key=itemgetter(2), reverse=self.reverse)
        # Then sort by mead-eval
        best_job = sorted(job, key=itemgetter(1), reverse=self.reverse)[0]
        # The result is sorted by mead-eval with ties broken by xpctl
        self.best_value = (best_job[1], best_job[2])
        return self._create_result_dict(best_job[0])


def main():  # pylint: disable=too-many-statements
    """Select a model for export if one meets the criteria
    """
    parser = argparse.ArgumentParser(description='Select a model for export if one meets the criteria')
    parser.add_argument('--cred', help='cred file', default="/etc/odind/odin-cred.yml")
    parser.add_argument('--type', help='Policy type', required=True)
    parser.add_argument(
        '--label', required=True, help="The odin task label for this selecting task, used to access the store"
    )
    parser.add_argument('--models', required=True, nargs='+')
    parser.add_argument('--dataset', help="(deprecated) The name of the dataset to evaluate", required=False)
    parser.add_argument('--task', required=False)
    parser.add_argument('--metric', default='acc')
    parser.add_argument('--user_cmp', default=None)
    parser.add_argument('--config', help='(deprecated) JSON Configuration for an experiment', type=convert_path)
    parser.add_argument(
        '--settings',
        help='JSON Configuration for mead',
        required=False,
        default='config/mead-settings.json',
        type=convert_path,
    )
    parser.add_argument('--datasets', help='(deprecated) json library of dataset labels', type=convert_path)
    parser.add_argument('--logging', help='json file for logging', default='config/logging.json', type=convert_path)
    parser.add_argument('--data_root', help='Data directory', default='/data')
    parser.add_argument('--xpctl_api_url', help='XPCTL api', type=str)

    args = parser.parse_args()

    if args.datasets is not None:
        LOGGER.warning("--datasets is unused and unneeded for calls to `odin-select`")
    if args.config is not None:
        LOGGER.warning("--config is unused and unneeded for calls to `odin-select`")
    if args.dataset is not None:
        LOGGER.warning("--dataset is unused and unneeded for calls to `odin-select`")

    cred_params = read_config_stream(args.cred)

    store = create_store_backend(**cred_params['jobs_db'])
    args.store = store

    xpctl_url = args.xpctl_api_url if args.xpctl_api_url is not None else cred_params['reporting_db']['host']
    args.api = xpctl_client(host=xpctl_url)

    params = vars(args)
    del params['cred']
    policy = create_export_policy(args.type, params)
    results = policy.select(args.models)
    if results:
        print(results)

        job_details = store.get(args.label)
        outputs = job_details.get("outputs", {})
        if outputs is None:
            outputs = {}
            job_details['outputs'] = outputs
        outputs.update(results)
        store.set(job_details)


if __name__ == "__main__":
    main()



"""Client to authenticate a user and cache a new JWT token"""
import argparse
from getpass import getuser
import os
from odin import LOGGER
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME
from odin.utils.auth import get_jwt_token


def authenticate_user(url: str, token_path: str, username: str, password: str) -> None:
    """Authenticate a user over HTTP
    :param url: the base URL
    :param token_path: The file location of the JWT token
    :param username: The user ID
    :param password: The password
    """
    if os.path.exists(token_path):
        os.remove(token_path)
    jwt_token = get_jwt_token(url, token_path, username, password)
    LOGGER.info(jwt_token)


def main():
    """Authenticate a user and give back a JWT token
    """
    parser = argparse.ArgumentParser(description='Create or update an odin user.  Requires valid JWT token to run')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument('--token', help="File where JWT token can reside", default=os.path.expanduser("~/.odin.token"))
    parser.add_argument('--username', '-u', help="Create or update a username", default=getuser())
    parser.add_argument('--password', '-p', help="New or updated password")
    parser.add_argument('--scheme', choices={'http', 'https'}, default=ODIN_SCHEME, help='The protocol to communicate over')
    args = parser.parse_args()
    url = f'{args.scheme}://{args.host}:{args.port}'
    authenticate_user(url, args.token, args.username, args.password)


if __name__ == '__main__':
    main()



"""A websocket client to get or stream logs from a container."""

import json
import asyncio
import argparse
import signal
from typing import Optional
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient


async def request_logs(
    ws: str, resource: str, namespace: str, container: str, follow: bool, lines: Optional[int] = None
) -> None:
    """Make the websocket request to get the logs.

    :param ws: The websocket host
    :param resource: The thing to get logs from
    :param namespace: The namespace the pod is in
    :param container: The container to get the logs from, only needed
        when there are multiple containers in a pod
    :param follow: Should you get all the logs available right now or stream
        them as they come in?
    :param lines: How many lines to grab

    """
    work = {'resource': resource, 'namespace': namespace, 'follow': follow}
    if container is not None:
        work['container'] = container
    if lines is not None:
        work['lines'] = lines
    async with websockets.connect(ws) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'LOGS', APIField.REQUEST: work}))
        line = json.loads(await websocket.recv())
        while line[APIField.STATUS] != APIStatus.END:
            if line[APIField.STATUS] == APIStatus.ERROR:
                LOGGER.error(line)
                break
            LOGGER.info(line[APIField.RESPONSE])
            line = json.loads(await websocket.recv())


def log_all_children_http(url, resource, namespace):
    client = HttpClient(url=url)
    try:
        print(client.request_logs(resource, namespace))
    except:
        head = find_head(resource)
        data = client.request_data(head)
        for job in data['jobs']['jobs']:
            rid = client.request_data(job)['jobs']['resource_id']
            print('================')
            print(rid)
            print('----------------')
            try:
                logs = client.request_logs(rid, namespace)
                print(logs)
            except:
                try:
                    # If everything still fails, it could be a Kubeflow job child
                    logs = client.request_logs(f'{rid}-master-0', namespace)
                    print(logs)
                except:
                    print('Failed to get log')


def find_head(resource):
    # Lets try and find this resource
    head = resource.split('j--')[0]
    if not head.endswith('j'):
        head += 'j'
    return head


def main():
    """Async websocket client to get logs from k8s pods."""

    signal.signal(signal.SIGINT, lambda *args, **kwargs: exit(0))

    parser = argparse.ArgumentParser()
    parser.add_argument('resource', help="The name of pod to get logs from.")
    parser.add_argument('--namespace', default='default', help="The namespace the pod is in.")
    parser.add_argument(
        '--container', help="The container to get logs from, only needed if there are multiple containers."
    )
    parser.add_argument(
        '--follow', '-f', action='store_true', help="Should you stream the logs or just get what is there now?"
    )
    parser.add_argument('--lines', '-l', type=int, help="How many lines from the end of the logs to grab")
    parser.add_argument('--host', default=ODIN_URL, type=str, help="The location of the server")
    parser.add_argument('--port', default=ODIN_PORT, help="The port the server listens on.")
    parser.add_argument(
        '--scheme',
        choices={'http', 'https', 'wss', 'ws'},
        default=ODIN_SCHEME,
        help='Websocket connection protocol, use `wss` for remote connections and `ws` for localhost',
    )
    args = parser.parse_args()

    endpoint = f'{args.scheme}://{args.host}:{args.port}'

    if args.scheme.startswith('http'):
        log_all_children_http(endpoint, args.resource, args.namespace)

    else:

        asyncio.get_event_loop().run_until_complete(
            request_logs(endpoint, args.resource, args.namespace, args.container, args.follow, args.lines)
        )


if __name__ == "__main__":
    main()



"""DSL to execute common chores
"""
import argparse
import logging
import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import List, Dict, Callable, NewType, Optional
import git

from baseline.utils import exporter, optional_params, read_config_stream, import_user_module
from odin.core import create_graph, _to_kwargs, wire_inputs, format_output
from odin.dag import topo_sort, find_children, dot_graph
from odin.store import create_store_backend

__all__ = []
export = exporter(__all__)
CHORE_REGISTRY = {}
Chore = NewType('Chore', Callable)

DEPENDENCY_KEY = 'depends'
LOGGER = logging.getLogger('odin')


def set_repo_creds(
    repo: git.Repo,
    name: Optional[str] = None,
    email: Optional[str] = None,
    name_env: str = "ODIN_GIT_NAME",
    email_env: str = "ODIN_GIT_EMAIL",
) -> git.Repo:
    """Set the name and email on the git repo so you can commit.

    :param repo: The repo object
    :param name: The name to set
    :param email: The email to set
    :param name_env: The env variable to read the name from if no name provided
    :param email_env: The env variable to read the email from if no email is provided

    :returns: The repo with the creds set
    """
    name = os.environ[name_env] if name is None else name
    email = os.environ[email_env] if email is None else email
    repo.config_writer().set_value("user", "name", name).release()
    repo.config_writer().set_value("user", "email", email).release()
    return repo


class ChoreSkippedException(Exception):
    """A special exception chores use to signal they want to quit."""


@contextmanager
def change_dir(path: str):  # pylint: disable=missing-yield-doc,missing-yield-type-doc
    """Change the current directory

    :param path: Change to what
    """
    current_dir = os.getcwd()
    try:
        os.chdir(os.path.expanduser(path))
        yield
    finally:
        os.chdir(current_dir)


@export
@optional_params
def register_chore(chore: Chore, name: str = None) -> Chore:
    """Register a chore (some callable) to a name
    :param chore: Some callable
    :param name: A unique name for this registry
    :raises Exception: If the name is already registered.
    :return: The callable
    """
    name = name if name is not None else chore.__name__
    if name in CHORE_REGISTRY:
        raise Exception(
            'Error: attempt to re-define previously'
            f' registered Chore {name} (old: {CHORE_REGISTRY[name]}, new: {chore}) in registry'
        )
    CHORE_REGISTRY[name] = chore
    return chore


@export
def create_chore(name: str) -> Chore:
    """Factory method to give back a chore requested by name

    :param name: Name of the chore
    :return: The (callable) Chore
    """
    return CHORE_REGISTRY[name]


@export
@register_chore('copy')
def copy(src: str, dst: str, clobber: bool = True) -> str:
    """Copy file from source to destination

    :param src: source path
    :param dst: destination path
    :param clobber: Should we overwrite (defaults to `True`)
    :raises ChoreSkippedException: If there is missing inputs.
    :return: The destination path
    """
    if src is None or dst is None:
        raise ChoreSkippedException()
    src = os.path.abspath(os.path.expanduser(src))
    dst = os.path.abspath(os.path.expanduser(dst))

    if os.path.isfile(src):
        # If dst is a directory that exists then we make a file in it
        # Not sure if we want this yet, only use cause rn is for dirs
        if os.path.exists(dst) and os.path.isdir(dst):
            dst = os.path.join(dst, os.path.basename(src))
        shutil.copy2(src, dst)
    else:
        dst = os.path.join(dst, os.path.basename(src))
        if clobber and os.path.exists(dst):
            if os.path.isdir(dst):
                shutil.rmtree(dst)
            else:
                os.remove(dst)
        shutil.copytree(src, dst)
    return dst


@export
@register_chore('git-add')
def git_add(dir: Path, files: List[str] = None) -> bool:
    """Add a directory to git with an optional list of files

    :param dir: A directory
    :param files: Files (defaults to `None`)
    :raises ChoreSkippedException: If there is missing inputs.
    :return: `True`
    """
    from git import Repo  # pylint: disable=import-outside-toplevel

    if files is None or any(f is None for f in files):
        raise ChoreSkippedException()
    if dir is not None:
        repo = Repo(dir)
    else:
        repo = Repo(files[0], search_parent_directories=True)
    repo = set_repo_creds(repo)
    for filename in files:
        repo.git.add(os.path.expanduser(filename))
    return True


@export
@register_chore('git-commit')
def git_commit(dir: Path, message: str) -> bool:
    """Commit a directory to git

    :param dir: The directory
    :param message: The git message
    :raises ChoreSkippedException: If there is missing inputs.
    :return: `True`
    """
    from git import Repo  # pylint: disable=import-outside-toplevel

    if dir is None or message is None:
        raise ChoreSkippedException()
    repo = Repo(dir)
    repo = set_repo_creds(repo)
    # Check that files have been staged for commit
    if not repo.index.diff("HEAD"):
        raise ChoreSkippedException()
    repo.git.commit(m=message)
    return True


def _repo_version_match(repo: git.Repo) -> bool:
    master = str(git.repo.fun.rev_parse(repo, 'origin/master'))
    head = str(git.repo.fun.rev_parse(repo, 'HEAD'))
    return master == head


def _stash_pull_pop(repo: git.Repo) -> None:
    do_stash = False
    if repo.is_dirty():
        do_stash = True
        repo.git.stash()
    if not _repo_version_match(repo):
        repo.git.pull('--rebase', 'origin', 'master')
    if do_stash:
        repo.git.stash('pop')


@export
@register_chore('git-push')
def git_push(dir: Path) -> bool:
    """Push a directory to git

    :param dir: Directory
    :raises ChoreSkippedException: If there is missing inputs.
    :return: `True`
    """
    from git import Repo  # pylint: disable=import-outside-toplevel

    if dir is None:
        raise ChoreSkippedException()
    repo = Repo(dir)
    repo = set_repo_creds(repo)
    _stash_pull_pop(repo)
    repo.git.push()
    return True


@export
@register_chore('git-pull')
def git_pull(dir: Path) -> bool:
    """Pull a git repo.

    :param dir: The directory where the git repo lives
    :raises ChoreSkippedException: If there is missing inputs.
    :returns: True
    """
    from git import Repo  # pylint: disable=import-outside-toplevel

    if dir is None:
        raise ChoreSkippedException()
    repo = Repo(dir)
    repo = set_repo_creds(repo)
    repo.git.pull()
    return True


@export
@register_chore('bump-version')
def bump_version(file_name: Path) -> str:
    """Bump the version with the ruamel round trip loader to preserve formatting."""
    import ruamel.yaml  # pylint: disable=import-outside-toplevel

    if file_name is None:
        raise ChoreSkippedException()
    with open(file_name) as rf:
        config = ruamel.yaml.load(rf, Loader=ruamel.yaml.RoundTripLoader)
    metadata = config.setdefault('metadata', {})
    labels = metadata.setdefault('labels', {})
    labels['version'] = str(int(labels.get('version', 0)) + 1)
    with open(file_name, 'w') as out:
        ruamel.yaml.dump(config, out, Dumper=ruamel.yaml.RoundTripDumper)
    return file_name


@export
@register_chore('bump-k8s')
def bump_k8s(base_dir: Path, deployment: str, client: str) -> List[str]:
    """Bump the version in the deployment and service yaml.

    :param base_dir: `str` The root path of all the orchestration files.
    :param deployment: `str` which deployment to bump, research, local, etc
    :param client: `str` which client to bump.
    :raises ChoreSkippedException: If there is missing inputs.

    :returns: List[str] A list of all files updated
    """
    if base_dir is None or deployment is None or client is None:
        raise ChoreSkippedException()
    types = ('deployments', 'services')
    return [bump_version(os.path.join(base_dir, deployment, kind, f"{client}.yml")) for kind in types]


@export
@register_chore('slack-message')
def slack_webhook(parent_details: Dict, webhook: str, template: str) -> None:
    """Substitute a template message and post to slack

    :param parent_details: The context to use to replace values in the template.
    :param webhook: The webhook key
    :param template: The message.
    """
    import requests  # pylint: disable=import-outside-toplevel
    from string import Template  # pylint: disable=import-outside-toplevel

    message: str = Template(template).substitute(parent_details)
    requests.post(webhook, json={"text": message})


def run_chores(config: Dict, results: Dict) -> Dict:
    """Execute a chore graph

    :param config: The list of chores and there deps
    :param results: A set of upstream results that can be used
    :return: The new results, updated with chore output
    """
    graph = create_graph(config['chores'], results)

    # Convert the graph from indices to names so it prints better.
    named_graph = {config['chores'][k]['name']: [config['chores'][v]['name'] for v in vs] for k, vs in graph.items()}
    LOGGER.info(dot_graph(named_graph))

    execution_order = topo_sort(graph)
    # Track nodes that give up and all the nodes that are downstream of them.
    to_skip = set()
    children = find_children(graph)
    LOGGER.info("Topological Sort %s", execution_order)
    for step in execution_order:
        params = config['chores'][step]
        name = params.pop('name')
        if step in to_skip:
            # If an ancestor of yours gave up you should give up.
            LOGGER.debug("Chore %s skipped because of parent", name)
            continue
        chore_type = params.pop('type')
        params.pop(DEPENDENCY_KEY, None)
        LOGGER.debug('Feeding Inputs: %s(%s)', name, _to_kwargs(params))
        LOGGER.debug(results)
        chore = create_chore(chore_type)
        params = wire_inputs(params, results, chore)
        LOGGER.info('Running: %s(%s)', name, _to_kwargs(params))
        try:
            res = chore(**params)
        except ChoreSkippedException:
            # I decided to give up. All my descendants should give up.
            res = None
            LOGGER.debug("Chore %s raised a ChoreSkippedException", name)
            to_skip.add(step)
            to_skip.update(children[step])
        results[name] = format_output(res)
    LOGGER.info("Chores %s skipped running.", to_skip)
    return results


def main():
    """Driver program to execute chores."""
    parser = argparse.ArgumentParser(description='Run chores')
    parser.add_argument('file', help='A chore YAML file.')
    parser.add_argument('--cred', help='cred file', default="/etc/odind/odin-cred.yml")
    parser.add_argument('--label', required=True)
    parser.add_argument('--modules', nargs='+', default=[], help='Addon modules to load')
    args = parser.parse_args()

    for addon in args.modules:
        import_user_module(addon)

    cred_params = read_config_stream(args.cred)
    store = create_store_backend(**cred_params['jobs_db'])
    config = read_config_stream(args.file)
    previous = store.get_previous(args.label)
    parent_details = store.get_parent(args.label)
    results = {prev_job_details['name']: prev_job_details for prev_job_details in previous}
    results['parent'] = parent_details
    results = run_chores(config, results)
    results = {'chore_context': results}
    LOGGER.info(results)
    job_details = store.get(args.label)
    job_details.update(results)
    store.set(job_details)


if __name__ == "__main__":
    main()



"""Tools to hash files."""

import os
from pathlib import Path
from hashlib import sha1
from itertools import chain
from functools import partial
from typing import List, Union, BinaryIO, Any, Optional
from baseline.utils import str_file, listify
from odin import LOGGER


# There is no hashlib base class so create something for typing.
Hash = Any  # pylint: disable=invalid-name


@str_file(f='rb')
def hash_file(f: Union[str, BinaryIO], hasher: Hash, block_size: int = 1024) -> Hash:
    """Update a hash function with the contents for a file.

    :param f: The file to hash.
    :param hasher: The hash object to use.
    :param block_size: The size of chunks used to read the file in. This
        doesn't effect the output of the hash function.

    :returns: The hash object updated with the file contents.
    """
    for block in iter(partial(f.read, block_size), b''):
        hasher.update(block)
    return hasher


def hash_files(files: Union[str, List[str]], hasher: Optional[Hash] = None, block_size: int = 1024) -> str:
    """Hash a list of files.

    :param files: The list of files to hash.
    :param hasher: The hash object. defaults to sha1.
    :param block_size: The size of chunks used to read the files in.

    :returns: The hexdigest of all the files.
    """
    hasher = hasher if hasher is not None else sha1()
    for f in sorted(expand_dirs(listify(files))):
        hasher = hash_file(f, hasher, block_size)
    return hasher.hexdigest()


def expand_dirs(files: List[str]) -> List[str]:
    """Given a list of files and dirs return a list all files in the dir.

    :param files: The list of files and dirs.

    :returns: The list with dirs expanded into the files contained within them.
    """
    new_files = []
    for f in files:
        f = os.path.expanduser(f)
        if not os.path.exists(f):
            LOGGER.warning("Requested hash of %s but file not found.", f)
            continue
        if os.path.isdir(f):
            new_files.extend(expand_dir(f))
        else:
            new_files.append(f)
    return new_files


def expand_dir(dir_name: str) -> List[str]:
    """Get a list of files inside a directory.

    :param dir_name: The dir to find files in.

    :returns: The list of files.
    """
    return map(str, filter(lambda x: not os.path.isdir(x), Path(dir_name).rglob("*")))


def hash_args(command: Union[str, List[str]], args: List[str]) -> str:
    """Hash the command and arguments of a container.

    :param command: The command that k8s will give the container.
    :param args: The arguments the container receives.
    :returns: The hash of the container args.
    """
    return sha1("".join(chain(listify(command), args)).encode('utf-8')).hexdigest()



"""Formatting utils
"""
from itertools import chain
from typing import Set, Optional
from collections import namedtuple


# https://stackoverflow.com/questions/5909873/how-can-i-pretty-print-ascii-tables-with-python
def print_table(rows: namedtuple, columns: Optional[Set[str]] = None) -> None:
    """Pretty print a table

    :param rows: Some rows
    :param columns: A set of columns to include in the output.
    """
    columns = columns if columns is not None else set(rows[0]._fields)
    fields = [f for f in rows[0]._fields if f in columns]
    if len(rows) > 1:
        lens = [
            len(str(max(chain((getattr(x, field) for x in rows), [field]), key=lambda x: len(str(x)))))
            for field in fields
        ]
        formats = []
        hformats = []
        for i, field in enumerate(fields):
            if isinstance(getattr(rows[0], field), int):
                formats.append("%%%dd" % lens[i])
            else:
                formats.append("%%-%ds" % lens[i])
            hformats.append("%%-%ds" % lens[i])
        pattern = " | ".join(formats)
        hpattern = " | ".join(hformats)
        separator = "-+-".join(['-' * n for n in lens])
        print(hpattern % tuple(fields))
        print(separator)
        for line in rows:
            print(pattern % tuple(getattr(line, field) for field in fields))
    elif len(rows) == 1:
        row = rows[0]
        hwidth = len(max(fields, key=lambda x: len(x)))
        for field in fields:
            print("%*s = %s" % (hwidth, field, getattr(row, field)))



"""Pipeline primitive
"""
import asyncio
from datetime import datetime
import json
import logging
from hashlib import sha1
from itertools import chain
from typing import Dict, List, Optional, Union, AsyncIterator, Any
from baseline.utils import listify
from mead.utils import order_json
from odin.core import create_graph, is_reference, parse_reference, extract_outputs, _get_child_name
from odin.dag import topo_sort_parallel, dot_graph, CycleError
from odin.k8s import Task, TaskManager, KubernetesTaskManager, StatusType, SubmitError
from odin.store import Store, Cache, MemoryCache
from odin.utils.hash import hash_files, hash_args

LOGGER = logging.getLogger('odin')


def hash_outputs(outputs: Dict[str, Union[str, List[str]]]) -> str:
    """Hash the outputs of a task.

    :param outputs: The list of output files to hash.
    :returns: The hash of the outputs.
    """
    LOGGER.debug("Hashing %s", outputs)
    output_hashes = {k: hash_files(listify(v)) for k, v in outputs.items()}
    LOGGER.debug(output_hashes)
    out_hash = sha1(json.dumps(order_json(output_hashes)).encode('utf-8')).hexdigest()
    LOGGER.debug("Output hash: %s", out_hash)
    return out_hash


async def hash_inputs(task: Task, sched: TaskManager) -> str:
    """Hash all things that are inputs to a task, the containers, the arguments, and input data.

    :param task: The Task that will be run.
    :param sched: The scheduler, needed to get the container hashes.
    :returns: The hash that describes the inputs to the Task.
    """
    LOGGER.debug("Hashing inputs for %s", task.name)
    arg_hash = hash_args(task.command, task.args)
    LOGGER.debug("Argument hash: %s", arg_hash)
    container_hash = await sched.hash_task(task)
    LOGGER.debug("Container hash: %s", container_hash)
    input_hash = hash_files(task.inputs) if task.inputs is not None else ""
    LOGGER.debug("Input data hash: %s", input_hash)
    full_hash = sha1(f"{arg_hash}{''.join(container_hash)}{input_hash}".encode('utf-8')).hexdigest()
    LOGGER.debug("Full input hash: %s", full_hash)
    return full_hash


class PipelineStatus:
    """Enum of pipeline status.

    Not a real enum to get easier access to string values.
    """

    BUILDING = "BUILDING"
    RUNNING = "RUNNING"
    TERMINATED = "TERMINATED"
    DONE = "DONE"


class Executor:
    """A task scheduler backed by kubernetes

    Pipelines are long-running tasks and therefore implemented with
    `asyncio` and `async/await`.  There are 2 components
    that an `Executor` manages: the tasks `Store` and the `TaskManager`

    Tasks that have no dependencies with each other are scheduled to be run in parallel
    """

    def __init__(
        self, store: Store, sched: TaskManager = None, cache: Optional[Cache] = None, modules: List[str] = None
    ):
        """Create a Pipeline, inject params

        :param store: The `Store` provides access to the job DB
        :param sched: The `TaskManager` handles Task management
        :param cache: The key value store used to track input and output hashes for tasks.
        """
        self.sched = sched if sched else KubernetesTaskManager(store, modules=modules)
        self.store = store
        self.cache = cache if cache else MemoryCache()

    @staticmethod
    def _task_to_entry(my_id: str, child_task_id: str, name: str, task: Task) -> object:
        return {
            Store.PIPE_ID: child_task_id,
            Store.PARENT: my_id,
            "command": task.command,
            "name": name,
            "image": task.image,
            "args": task.args,
            "resource_type": task.resource_type,
            "node_selector": task.node_selector,
            "pull_policy": task.pull_policy,
            "num_gpus": task.num_gpus,
            "outputs": task.outputs,
            "inputs": task.inputs,
        }

    def extract_outputs(self, pipeline_id: str, reference: str) -> Any:
        """Pull data out of the job store based on the reference path.

        :param pipeline_id: The id of the pipeline that is currently running.
        :param reference: The reference in the form `^a.b` to extract data from.
        :returns: The data pulled out..
        """
        assert is_reference(reference)
        task, *path = parse_reference(reference)
        label = _get_child_name(pipeline_id, task)
        task_entry = self.store.get(label)
        return extract_outputs(path, task_entry)

    def rewrite_references(self, pipeline_id: str, reference: str) -> Any:
        """Pull data out of the job store to replace subsections of arguments.

        :param pipeline_id: The id of the pipelines that is currently running.
        :param reference: The argument with references in the form `^a.b` if you
            wrap these in curlies like `example-{^a.b}-more` you can replace just
            one part of it.
        :returns: The input with references replaced with the data from the job store.
        """
        starts = []
        new_ref = []
        for i, char in enumerate(reference):
            # If we are at the start of a possible reference span
            if char == "{":
                starts.append(i + 1)
            # We hit the end of a span
            elif char == "}":
                # If this is a } without a start it must just be something the user has in text so skip
                if not starts:
                    new_ref.append(char)
                    continue
                # Grab the opening of this span
                start = starts.pop()
                if not starts:
                    to_sub = reference[start:i]
                    # Recursively substitute in the open span
                    sub = self.rewrite_references(pipeline_id, reference[start:i])
                    # If this open span wasn't a real substitution add the `{` and `}` back
                    if sub == to_sub:
                        sub = f"{{{sub}}}"
                    # Save out the substituted value
                    new_ref.append(sub)
            else:
                # If outside of a span collect the text for later use
                if not starts:
                    new_ref.append(char)
        # If there is a span that is never closed copy out everything since then
        if starts:
            new_ref.append(reference[starts[0] - 1 :])
        # Combine the substituted sections into a single string
        reference = "".join(new_ref)
        # If the whole string is a reference just substitute it (base case)
        if is_reference(reference):
            return self.extract_outputs(pipeline_id, reference)
        # If the whole thing isn't a reference just return as is (with possible sub references subsituted)
        return reference

    async def run(  # pylint: disable=too-many-locals,too-many-statements,too-many-branches,missing-yield-type-doc
        self, my_id: str, task_id: str, rev_ver: str, tasks: List[Dict]
    ) -> AsyncIterator[str]:
        """Run a pipeline by stepping through sets of tasks

        This (async) method manages scheduling a series of
        tasks, launching each, updating the parent
        status in the jobs DB, and waiting for
        each task to complete.  A pipeline records its
        success or failure directly to the jobs DB

        :param my_id: The pipeline unique identifier
        :param task_id: The task unique identifier
        :param rev_ver: The revision version if available
        :param tasks: The tasks to execute (with their params)
        :raises ValueError: If the pipeline definition fails.
        :raises KeyError: If the pipeline definition fails.
        :return: AsyncIterator[str]
        """
        try:
            graph = create_graph(tasks)
            named_graph = {tasks[k]['name']: [tasks[v]['name'] for v in vs] for k, vs in graph.items()}
            LOGGER.info(dot_graph(named_graph))
            groups = topo_sort_parallel(graph)
        except (KeyError, ValueError, CycleError) as exc:
            self.store.set(
                {
                    Store.PIPE_ID: my_id,
                    Store.JOB_NAME: task_id,
                    Store.REV_VER: rev_ver,
                    Store.STATUS: PipelineStatus.TERMINATED,
                    Store.EXECUTED: [],
                    Store.WAITING: [],
                    Store.EXECUTING: [],
                    Store.SUBMIT_TIME: datetime.utcnow(),
                    Store.COMPLETION_TIME: None,
                    Store.ERROR_MESSAGE: str(exc),
                }
            )
            yield f"Pipeline {my_id} terminated"
            raise exc
        self.store.set(
            {
                Store.PIPE_ID: my_id,
                Store.JOB_NAME: task_id,
                Store.REV_VER: rev_ver,
                Store.STATUS: PipelineStatus.BUILDING,
                Store.EXECUTED: [],
                Store.WAITING: [],
                Store.EXECUTING: [],
                Store.SUBMIT_TIME: datetime.utcnow(),
                Store.COMPLETION_TIME: None,
                Store.ERROR_MESSAGE: None,
            }
        )
        task_list = []
        waiting = []

        for group in groups:
            task_group = []
            waiting_group = []
            for task in group:
                child_task_id = tasks[task]['name']
                task_obj = Task.from_dict(tasks[task])
                task_entry = Executor._task_to_entry(my_id, child_task_id, tasks[task]['_name'], task_obj)
                self.store.set(task_entry)
                task_group.append(task_obj)
                waiting_group.append(child_task_id)
            task_list.append(task_group)
            waiting.append(waiting_group)

        my_status = self.store.get(my_id)
        my_status['status'] = PipelineStatus.RUNNING
        all_tasks = list(chain(*waiting))
        my_status[Store.WAITING] = all_tasks
        my_status[Store.JOBS] = all_tasks
        self.store.set(my_status)

        for task_group in task_list:
            # We are about to move this first set from waiting to executing so remove it from waiting.
            waiting.pop(0)
            # Last rounds tasks are still in the EXECUTING state so set them to EXECUTED
            if my_status[Store.EXECUTING]:
                my_status[Store.EXECUTED].extend(my_status[Store.EXECUTING])
            my_status[Store.EXECUTING] = [task_obj.name for task_obj in task_group]
            my_status[Store.WAITING] = list(chain(*waiting))
            self.store.set(my_status)

            running = []
            for task_obj in task_group:
                # Replace inputs and args with values generated by previous tasks
                task_obj.inputs = (
                    [self.rewrite_references(my_id, task_input) for task_input in task_obj.inputs]
                    if task_obj.inputs is not None
                    else None
                )
                task_obj.args = [self.rewrite_references(my_id, a) for a in task_obj.args]
                task_entry = self.store.get(task_obj.name)
                task_entry['args'] = task_obj.args
                task_entry['inputs'] = task_obj.inputs
                self.store.set(task_entry)

                if task_obj.outputs is not None:
                    try:
                        in_hash = await hash_inputs(task_obj, self.sched)
                    except SubmitError as exc:
                        my_status[Store.STATUS] = PipelineStatus.TERMINATED
                        my_status[Store.ERROR_MESSAGE] = str(exc)
                        self.store.set(my_status)
                        yield f"Pipeline {my_id} terminated"
                        raise exc
                    out_hash = hash_outputs(task_obj.outputs)
                    prev_hash = self.cache[in_hash]
                    if out_hash == prev_hash:
                        LOGGER.info("%s is cached and will not be run", task_obj.name)
                        task_status = self.store.get(task_obj.name)
                        task_status.update({Store.RESOURCE_ID: Store.CACHED})
                        self.store.set(task_status)
                        my_status[Store.EXECUTING].remove(task_obj.name)
                        my_status[Store.EXECUTED].append(task_obj.name)
                        self.store.set(my_status)
                        continue
                    LOGGER.info("Hash of outputs for %s doesn't match stored hash, re-running.", task_obj.name)
                LOGGER.info("Submitting %s", task_obj.name)
                yield f"Submitting {task_obj.name}"
                try:
                    resource_id = self.sched.submit(task_obj)
                except SubmitError as exc:
                    my_status[Store.STATUS] = PipelineStatus.TERMINATED
                    my_status[Store.ERROR_MESSAGE] = str(exc)
                    self.store.set(my_status)
                    yield f"Pipeline {my_id} terminated"
                    raise exc

                task_status = self.store.get(task_obj.name)
                task_status.update({Store.RESOURCE_ID: resource_id, Store.SUBMIT_TIME: datetime.utcnow()})
                self.store.set(task_status)
                running.append(task_obj)

            # This processes jobs in the order they finish, we should eventually
            # be able to kill the other jobs if one gets terminated.
            for future in asyncio.as_completed(map(self.sched.wait_for, running)):
                task_obj = await future
                LOGGER.info("Done running %s", task_obj.name)
                yield f"Done running {task_obj.name}"
                task_status = self.store.get(task_obj.name)
                task_status[Store.COMPLETION_TIME] = datetime.utcnow()
                self.store.set(task_status)
                completion_status = self.sched.status(task_obj)

                if StatusType(completion_status.status_type) is not StatusType.SUCCEEDED:
                    my_status[Store.STATUS] = PipelineStatus.TERMINATED
                    my_status[Store.ERROR_MESSAGE] = (
                        completion_status.message
                        if completion_status.message is not None
                        else f"Task `{task_obj.name}` failed"
                    )
                    self.store.set(my_status)
                    LOGGER.error([my_status, completion_status.message])
                    continue

                if self.store.get(task_obj.name).get(Store.REQUEST_EARLY_EXIT, False):
                    LOGGER.info("%s requested an early exit. Pipeline will complete now.", task_obj.name)
                    my_status[Store.STATUS] = PipelineStatus.DONE

                my_status[Store.EXECUTING].remove(task_obj.name)
                my_status[Store.EXECUTED].append(task_obj.name)
                self.store.set(my_status)

                if task_obj.outputs is not None:
                    LOGGER.info("Saving the hash of %s's output", task_obj.name)
                    self.cache[in_hash] = hash_outputs(task_obj.outputs)
            if my_status[Store.STATUS] is PipelineStatus.DONE or my_status[Store.STATUS] is PipelineStatus.TERMINATED:
                if my_status[Store.STATUS] == PipelineStatus.TERMINATED:
                    yield f"Pipeline: {my_id} Terminated"
                waiting = []
                break

        my_status[Store.COMPLETION_TIME] = datetime.utcnow()
        if my_status[Store.EXECUTING] and my_status[Store.STATUS] != PipelineStatus.TERMINATED:
            my_status[Store.EXECUTED].extend(my_status[Store.EXECUTING])
            my_status[Store.EXECUTING] = []
        assert not waiting

        if my_status[Store.STATUS] != PipelineStatus.TERMINATED:
            my_status[Store.STATUS] = PipelineStatus.DONE
        self.store.set(my_status)



"""Define (partial) resource handler for k8s Deployments"""

from typing import List
from kubernetes import client
from odin.k8s import ResourceHandler, register_resource_handler


@register_resource_handler(aliases=['deploy'])
class DeploymentHandler(ResourceHandler):

    """(partial) resource handler for k8s Deployments"""

    NAME = "DEPLOYMENT"

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return DeploymentHandler.NAME

    def __init__(self, namespace: str):
        """Initialize from given namespace

        :param namespace: A given namespace
        :type namespace: str
        """

        super().__init__(namespace)
        self.api = client.AppsV1Api()

    def get_pods(self, name: str) -> List[str]:
        """Get the list of pods that are managed by a service.

        :param name: The name of the service you want the pods from.
        :returns: The pod names
        """

        try:
            scale = self.api.read_namespaced_deployment_scale(name, self.namespace)
            selectors = scale.status.selector
            return self.core_api.list_namespaced_pod(self.namespace, label_selector=selectors).items
        except client.rest.ApiException:
            return []

    def get_api(self):
        """Get the deployment API
        :return The deployment API
        """
        return self.api



"""Websocket client to get the status a job."""

import json
import asyncio
import argparse
from typing import Set
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.status import show_status, Row, Pipeline
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient


async def request_status(ws: str, work: str, columns: Set[str], all_cols: bool = False) -> None:
    """Request the status of an odin job over web-sockets
    :param ws: The web socket
    :param work: the job name
    :param columns: A set of columns to include in the output
    :param all_cols: Should we just show all columns, If true then columns in ignored
    """
    async with websockets.connect(ws) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'STATUS', APIField.REQUEST: work}))

        results = json.loads(await websocket.recv())
        if results[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(results)
            return
        if results[APIField.STATUS] == APIStatus.OK:
            results = results[APIField.RESPONSE]
            for result in results:
                rows = [Row(**r) for r in result['task_statuses']]
                show_status(Pipeline(**result['pipeline_status']), rows, columns, all_cols)


def _task2row(task):
    row = Row(
        task=task.get('task'),
        status=task.get('status'),
        command=task.get('command'),
        name=task.get('name'),
        image=task.get('image'),
        resource_type=task.get('resourceType'),
        resource_id=task.get('resourceId'),
        submitted=task.get('submitTime'),
        completed=task.get('completionTime'),
    )

    return row


def _result2pipe(result):
    pipe = Pipeline(
        label=result.get('id'),
        job=result.get('job'),
        submitted=result.get('submitTime'),
        completed=result.get('completionTime'),
        status=result.get('status'),
        version=result.get('version'),
    )

    return pipe


def request_status_http(url: str, work: str, columns: Set[str], all_cols: bool = False) -> None:
    """Request the status over HTTP
    :param url: the base URL
    :param work: The pipeline ID
    :param columns: A set of columns to include in the output
    :param all_cols: Should we just show all columns, If true then columns in ignored
    """
    results = HttpClient(url).request_status(work)
    for result in results:
        rows = [_task2row(r) for r in result['tasks']]
        show_status(_result2pipe(result), rows, columns, all_cols)


def main():
    """A websocket client to request a job status."""
    parser = argparse.ArgumentParser(description="Websocket-based job clean-up")
    parser.add_argument('work', help='Job')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws', 'http', 'https'},
        default=ODIN_SCHEME,
        help='Connection protocol, use `http` for REST, use `wss` for remote connections and `ws` for localhost',
    )
    parser.add_argument('--columns', nargs="+", default=[], help="Columns of the status to show.")
    parser.add_argument('--all', action='store_true', help="Show all columns of the status message.")
    args = parser.parse_args()

    url = f'{args.scheme}://{args.host}:{args.port}'

    if args.scheme.startswith('ws'):
        asyncio.get_event_loop().run_until_complete(request_status(url, args.work, set(args.columns), args.all))
    else:
        request_status_http(url, args.work, set(args.columns), args.all)


if __name__ == "__main__":
    main()



"""Websocket/HTTP client to get the status a job."""
from collections import namedtuple
import argparse
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient
from odin.utils.formatting import print_table

Row = namedtuple('Row', 'host gpu type free power util memuse processes pids ')


def _gpu2row(gpu, host):
    proc_group = gpu.get('processes', [])
    processes = ' '.join(p['process_name'] for p in proc_group)
    pids = ' '.join(str(p['pid']) for p in proc_group)
    free = 'YES' if len(proc_group) == 0 else 'NO'
    used_memory = sum([int(p['used_memory']) for p in proc_group])
    power = '{: >4}/{}{}'.format(int(gpu['powerReadings']['powerDraw']), int(gpu['powerReadings']['powerLimit']), gpu['powerReadings']['unit'])
    gpu_type = gpu['productName'].replace('GeForce ', '')
    util = '{}{}'.format(gpu['utilization']['gpuUtil'], gpu['utilization']['unit']) if gpu['utilization']['gpuUtil'] else ''
    gpu_id = gpu['id'].split(':')[1]
    row = Row(host=host, gpu=gpu_id, type=gpu_type, free=free, power=power, util=util, memuse='{}M'.format(used_memory) if used_memory else '', processes=processes, pids=pids)
    return row


def request_nodes_http(url: str) -> None:
    """Request the status over HTTP
    :param url: the base URL
    """
    nodes = HttpClient(url).request_cluster_hw_status()
    rows = []

    for node in nodes:
        for gpu in node['gpus']:
            rows.append(_gpu2row(gpu, node['host']))
    print_table(rows)


def main():
    """An HTTP client to request GPU status on a cluster."""
    parser = argparse.ArgumentParser(description="Get GPU status on a cluster over HTTP")
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument(
        '--scheme', choices={'http', 'https'}, default=ODIN_SCHEME, help='Connection protocol, supports HTTP and HTTPs',
    )
    args = parser.parse_args()

    url = f'{args.scheme}://{args.host}:{args.port}'
    request_nodes_http(url)


if __name__ == "__main__":
    main()



"""Defines a resource handler for Jobs"""

from kubernetes import client
from odin.store import Store
from odin.k8s import ResourceHandler, Task, Status, task_to_pod_spec, register_resource_handler


@register_resource_handler
class JobHandler(ResourceHandler):

    """resource handler for jobs"""

    NAME = "Job"

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return JobHandler.NAME

    def __init__(self, namespace: str):
        """Create a k8s Job handle with given namespace

        :param namespace: A namespace
        :type namespace: str
        """
        super().__init__(namespace)
        self.api = client.BatchV1Api()

    def get_api(self) -> object:
        """Get back the API for Jobs (BatchV1)

        :return: An API object
        :rtype: object
        """
        return self.api

    def submit(self, task: Task):
        """Submit a new task as a Job

        :param task: A task definition
        :type task: Task
        :return: A string identifier for this task
        :rtype: str
        """
        secrets = self._reference_secrets(task)
        configmaps = self._generate_configmaps(task)
        pod_spec = task_to_pod_spec(task, secrets=secrets, configmaps=configmaps)
        metadata = client.V1ObjectMeta(name=task.name)
        template_metadata = client.V1ObjectMeta(name='{}-template'.format(task.name))

        template = client.V1PodTemplateSpec(metadata=template_metadata, spec=pod_spec)

        task_spec = client.V1JobSpec(template=template)
        task_obj = client.V1Job(kind=JobHandler.NAME, spec=task_spec, metadata=metadata)

        self.api.create_namespaced_job(body=task_obj, namespace=self.namespace)
        return task.name

    def status(self, name: str, store: Store) -> Status:
        """Get back a status for this task

        :param name: The task name
        :type name: str
        :param store: The jobs store
        :type store: Store
        :return: A status for this task
        :rtype: Status
        """

        job_status = self.api.read_namespaced_job_status(name=name, namespace=self.namespace).status
        return job_status

    def kill(self, name, store: Store) -> None:
        """Kill a Job task with background cascaded delete on the pods

        https://github.com/kubernetes/kubernetes/issues/20902

        :param name: A job identifier
        :type name: str
        :param store: A job store
        :type store: Store
        :return: None
        """
        delete_options = client.V1DeleteOptions(api_version="batch/v1", propagation_policy="Background")
        return self.api.delete_namespaced_job(name=name, namespace=self.namespace, body=delete_options)

    def get_pods(self, name: str):
        """Get Job objects for this name

        Should we change to pods? We arent really using Jobs much RN

        :param name: A job identifier
        :type name: str
        :return: A list of jobs
        :rtype: List
        """
        try:
            results = self.api.read_namespaced_job(name, namespace=self.namespace)
            return [results]
        except Exception:
            return []



"""Websocket client to cleanup a job."""

import os
from getpass import getuser
import json
import asyncio
import argparse
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient
from odin.cleanup import Cleaned
from odin.utils.formatting import print_table
from odin.utils.auth import get_jwt_token


async def request_cleanup(ws: str, work: str, purge_db: bool = False, purge_fs: bool = False):
    """Request the work is cleaned up by the server."""
    async with websockets.connect(ws) as websocket:
        args = {'work': work, 'purge_db': purge_db, 'purge_fs': purge_fs}
        await websocket.send(json.dumps({APIField.COMMAND: 'CLEANUP', APIField.REQUEST: args}))

        results = json.loads(await websocket.recv())
        if results[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(results)
            return
        if results[APIField.STATUS] == APIStatus.OK:
            cleaned = results[APIField.RESPONSE]
            print("Results of this request:")
            print_table([Cleaned(**c) for c in cleaned])


def _result2cleanup(result):
    return Cleaned(
        task_id=result.get('taskId'),
        cleaned_from_k8s=result.get('cleanedFromK8s'),
        purged_from_db=result.get('purgedFromDb'),
        removed_from_fs=result.get('removedFromFs'),
    )


def request_cleanup_http(url: str, jwt_token: str, work: str, purge_db: bool = False, purge_fs: bool = False) -> None:
    """Request the status over HTTP
    :param url: the base URL
    :param jwt_token: The token for last user auth
    :param work: The pipeline ID
    :param purge_db: Should we delete the pipeline from the jobs db too?
    :param purge_fs: Should we remove pipeline file system artifacts?
    """
    results = HttpClient(url, jwt_token=jwt_token).delete_pipeline(work, purge_db, purge_fs)
    cleaned = [_result2cleanup(r) for r in results['cleanups']]
    print("Results of this request:")
    print_table(cleaned)


def main():
    """A websocket client to request a job cleanup."""
    parser = argparse.ArgumentParser(description="Websocket-based job clean-up")
    parser.add_argument('work', help='Job')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument('--token', help="File where JWT token can reside", default=os.path.expanduser("~/.odin.token"))
    parser.add_argument('--username', '-u', help="Username", default=getuser())
    parser.add_argument('--password', '-p', help="Password")
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws', 'http', 'https'},
        default=ODIN_SCHEME,
        help='Connection protocol, use `http` for REST, use `wss` for remote connections and `ws` for localhost',
    )
    parser.add_argument('--db', '-d', action='store_true', help="Also remove from the job db")
    parser.add_argument('--fs', '-f', action='store_true', help="Also remove from the filesystem")
    args = parser.parse_args()

    url = f'{args.scheme}://{args.host}:{args.port}'

    if args.scheme.startswith('ws'):
        asyncio.get_event_loop().run_until_complete(request_cleanup(url, args.work, args.db, args.fs))
    else:
        jwt_token = get_jwt_token(url, args.token, args.username, args.password)
        try:
            request_cleanup_http(url, jwt_token, args.work, args.db, args.fs)
        except ValueError:
            # Try deleting the token file and start again
            if os.path.exists(args.token):
                os.remove(args.token)
                jwt_token = get_jwt_token(url, args.token, args.username, args.password)
                request_cleanup_http(url, jwt_token, args.work, args.db, args.fs)


if __name__ == "__main__":
    main()



"""Tool to get odin job status
"""
import argparse
from collections import namedtuple
from typing import Union, List, Tuple, Set, Optional

from baseline.utils import exporter, listify, read_config_stream, color, Colors
from mead.utils import convert_path
from odin.store import Store, create_store_backend
from odin.executor import PipelineStatus
from odin.utils.formatting import print_table

__all__ = []
export = exporter(__all__)

DEFAULT_COLUMNS = {'task', 'status', 'command', 'resource_id', 'submitted'}
Row = namedtuple('Row', 'task status command name image resource_type resource_id submitted completed')
Pipeline = namedtuple('Pipeline', 'label job version status submitted completed')


def id2row(job_name: str, status: str, store: Store) -> Row:
    """Convert some id to a `Row`

    :param job_name: A job id
    :param status: the job status
    :param store: The jobs DB
    :return: A `Row`
    """
    job_details = store.get(job_name)
    return Row(
        job_name,
        status,
        job_details['command'],
        job_details['name'],
        job_details['image'],
        job_details.get('resource_type', "Pod"),
        job_details.get(Store.RESOURCE_ID),
        job_details.get(Store.SUBMIT_TIME),
        job_details.get(Store.COMPLETION_TIME),
    )


def ids2rows(job_names: Union[str, List[str]], status: str, store: Store) -> List[Row]:
    """Convert one or more ids a `List[Row]`

    :param job_names: A list
    :param status: the job status
    :param store: The jobs DB
    :return: A `List[Row]`
    """
    job_names = listify(job_names)
    return [id2row(job_name, status, store) for job_name in job_names]


def get_status(work: str, store: Store) -> Tuple[Pipeline, List[Row]]:
    """Get status for a pipeline

    :param work: A pipeline name
    :param store: A job store
    :return: The information about the pipeline and a list of information about each pod
    """
    parent_details = store.get(work)
    pipe = Pipeline(
        parent_details[Store.PIPE_ID],
        parent_details.get(Store.JOB_NAME),
        parent_details.get(Store.REV_VER),
        parent_details['status'],
        parent_details.get(Store.SUBMIT_TIME),
        parent_details.get(Store.COMPLETION_TIME),
    )
    rows = []
    job_names = parent_details[Store.EXECUTED]
    rows += ids2rows(job_names, Store.EXECUTED, store)

    job_name = parent_details[Store.EXECUTING]
    if job_name:
        rows += ids2rows(job_name, "terminated" if pipe.status == "TERMINATED" else Store.EXECUTING, store)

    job_names = parent_details[Store.WAITING]
    rows += ids2rows(job_names, Store.WAITING, store)
    return pipe, rows


def show_status(pipe: Pipeline, rows: List[Row], columns: Optional[Set[str]] = None, all_cols: bool = False) -> None:
    """Show the status for a pipeline.

    :param pipe: Information about the pipeline itself.
    :param rows: The rows of the table.
    :param columns: A set of columns to include in the output.
    :param all_cols: Include all the columns in the output.
    """
    status = None
    if pipe.status == PipelineStatus.DONE:
        status = color(pipe.status, Colors.CYAN)
    elif pipe.status == PipelineStatus.RUNNING:
        status = color(pipe.status, Colors.GREEN)
    elif pipe.status == PipelineStatus.TERMINATED:
        status = color(pipe.status, Colors.RED)
    elif pipe.status == PipelineStatus.BUILDING:
        status = color(pipe.status, Colors.YELLOW)
    width = max(len(pipe.label), len('Finished'))
    print(f'{pipe.label:<{width}} --> {status}')
    if pipe.submitted is not None:
        start = "Started"
        print(f'{start:<{width}} --> {pipe.submitted}')
    if pipe.completed is not None:
        fin = "Finished"
        print(f'{fin:<{width}} --> {pipe.completed}')
    print()
    if columns:
        columns.update(DEFAULT_COLUMNS)
    else:
        columns = DEFAULT_COLUMNS
    if rows:
        if all_cols:
            columns.update(rows[0]._fields)
        print_table(rows, columns)


def main():
    """Take in a job and get back its status

    TODO: support passing in specific Job IDs and regex
    """
    parser = argparse.ArgumentParser(description='Get job status')
    parser.add_argument('work', help='Pipeline or Job')
    parser.add_argument('--cred', help='cred file', type=convert_path, required=True)
    parser.add_argument('--format', help='Format the output', default="human")
    parser.add_argument('--columns', nargs="+", default=[], help="Columns of the status to show.")
    parser.add_argument('--all', action='store_true', help="Show all columns of the status message.")
    args = parser.parse_args()
    cred_params = read_config_stream(args.cred)
    store = create_store_backend(**cred_params['jobs_db'])
    work = store.parents_like(args.work)
    if not work:
        print('No job found')
    for parent in work:
        try:
            show_status(*get_status(parent, store), columns=set(args.columns), all_cols=args.all)
        except Exception:
            print('ERROR: Skipping {}'.format(parent))


if __name__ == "__main__":
    main()



from typing import Optional, List, Any
import sqlalchemy as sql
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.orm import relationship

from odin.store import Cache, Store, Dict, register_cache_backend, register_store_backend
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from datetime import datetime
from copy import deepcopy


class DB:

    SessionFactory = None
    Engine = None
    Base = declarative_base()

    @staticmethod
    def get_session_factory(pool_size=10, **kwargs):
        uri = kwargs.get('uri', None)
        if uri is None:
            dbtype = kwargs.get('dbtype', 'postgresql')
            host = kwargs.get('host', 'localhost')
            port = kwargs.get('port', None)
            if port is not None:
                host = '{}:{}'.format(host, port)
            username = kwargs.get('user', None)
            passwd = kwargs.get('passwd', None)

            user = username if passwd is None else '{}:{}'.format(username, passwd)
            dbname = kwargs.get('db', 'jobs_db')
            uri = '{}://{}@{}/{}'.format(dbtype, user, host, dbname)
        if not DB.SessionFactory:
            DB.Engine = sql.create_engine(uri, echo=False, paramstyle='format', pool_size=pool_size)
            DB.Base.metadata.create_all(DB.Engine)
            DB.SessionFactory = sessionmaker(DB.Engine)
        return DB.SessionFactory


class CacheTable(DB.Base):
    __tablename__ = 'hashes'
    id = sql.Column(sql.Integer, primary_key=True)
    inputs = sql.Column(sql.String, nullable=False)
    outputs = sql.Column(JSON, nullable=False)


class Job(DB.Base):
    __tablename__ = Store.JOBS
    id = sql.Column(sql.Integer, primary_key=True)
    label = sql.Column(sql.String, nullable=False)
    job_name = sql.Column(sql.String)
    version = sql.Column(sql.String)
    status = sql.Column(sql.String)
    submit_time = sql.Column(sql.DateTime)
    completion_time = sql.Column(sql.DateTime, nullable=True)
    parent_label = sql.Column(sql.String, nullable=True, default=None)
    details = sql.Column(JSON, nullable=False)


@register_cache_backend('postgres')
class PostgresCache(Cache):
    """A key-value cache backed by PG"""

    COLL = 'hashes'
    IN = 'inputs'
    OUT = 'outputs'
    ID = '_id'

    def __init__(
        self,
        host: str,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
        db: str = 'jobs_db',
        port: int = 5432,
        **kwargs,
    ):
        """Connect to the mongodb backend.

        :param host: The location of the db.
        :param user: The username to log in with.
        :param passwd: The password to use.
        :param db: The database to use.
        :param port: The port to connect on.
        :raises Exception: If connection to mongo fails.
        :raises ServerSelectionTimeoutError: If the connection times out.
        """
        super().__init__()
        self.Session = DB.get_session_factory(user=user, passwd=passwd, host=host, db=db, port=port)

    def __setitem__(self, key: str, value: Any) -> None:
        """Add to the cache, if the key exists overwrite it.

        :param key: The key to add to the cache.
        :param value: The value that goes with the key
        """
        session = self.Session()

        kv = session.query(CacheTable).filter(CacheTable.key == key).scalar()
        if kv:
            kv.value = value
        else:
            session.add(CacheTable(inputs=key, value=value))
        session.commit()

    def __getitem__(self, key: str) -> Any:
        """Get data from the cache. Returns None if key is missing.

        :param key: The key to look for.
        :returns: The value associated with the key.
        """
        session = self.Session()
        kv = session.query(CacheTable).filter(CacheTable.key == key)
        return kv.outputs

    def __delitem__(self, key: str) -> None:
        """Delete a key-value pair from the cache.

        :param key: The key to remove.
        """
        session = self.Session()

        kv = session.query(CacheTable).filter(CacheTable.key == key)
        if kv:
            session.delete(kv)

    def keys(self) -> List[str]:
        """Get the keys in the cache."""
        session = self.Session()
        stmt = sql.text("SELECT key FROM hashes")
        stmt = stmt.columns(CacheTable.inputs)
        return session.query(CacheTable.inputs).from_statement(stmt).all()


@register_store_backend('postgres')
class PostgresStore(Store):
    """MongoDB backed `Store` implementation

    This implementation fulfills the `Store` interface backed by a MongoDB
    """

    def __init__(
        self,
        host: str,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
        db: str = 'jobs_db',
        port: int = 5432,
        **kwargs,
    ):
        """Connect to the mongodb backend.

        :param host: The location of the db.
        :param user: The username to log in with.
        :param passwd: The password to use.
        :param db: The database to use.
        :param port: The port to connect on.
        :raises Exception: If connection to mongo fails.
        :raises ServerSelectionTimeoutError: If the connection times out.
        """
        super().__init__()
        self.Session = DB.get_session_factory(user=user, passwd=passwd, host=host, db=db, port=port)

    def _get_if(self, job_id: str, session=None):
        if session is None:
            session = self.Session()
        return session.query(Job).filter(Job.label == job_id)

    def get(self, job_id: str) -> Dict:
        """This gives back the result of the job store for this entry

        :param job_id: This is a unique ID for this job
        :raises KeyError: If the job is not in the database.
        :return: A dictionary containing the output user data
        """
        result = self._get_if(job_id).scalar()
        if result is None:
            raise KeyError(f"No job {job_id} found in jobs DB")
        job_info = self._pack_job_info(result)
        return job_info

    def _pack_job_info(self, job: Job) -> Dict:
        job_info = {}
        job_info[Store.PIPE_ID] = job.label
        job_info[Store.JOB_NAME] = job.job_name
        job_info['status'] = job.status
        job_info[Store.SUBMIT_TIME] = job.submit_time
        job_info[Store.REV_VER] = job.version
        job_info[Store.PARENT] = job.parent_label
        job_info.update(job.details)
        return job_info

    def get_parent(self, job_str: str) -> Dict:
        """Get job results from the parent job

        :param job_str: This job
        :return: The parent job
        """
        return self._get_if(job_str).first().parent

    def get_previous(self, job_str: str) -> List[Dict]:
        """Get all job results that can before this job

        :param job_str: This job
        :return: All previous
        """
        session = self.Session()
        parent_label = self._get_if(job_str).first().parent_label
        details = session.query(Job).filter(Job.parent_label == parent_label).first().details
        executed = details[Store.EXECUTED]
        return session.query(Job).label.in_(executed)

    def _set(self, value: Dict) -> None:
        """This updates the job store for this entry

        :param value:
        """

        d = deepcopy(value)
        id = d.pop(Store.PIPE_ID)
        parent_label = d.pop(Store.PARENT, None)
        job_name = d.pop(Store.JOB_NAME, None)
        completion_time = d.pop(Store.COMPLETION_TIME, None)
        submit_time = d.pop(Store.SUBMIT_TIME, datetime.now())
        status = d.pop("status", None)
        version = d.pop(Store.REV_VER, None)

        session = self.Session()
        kv = self._get_if(id, session).scalar()
        if kv is not None:
            kv.details = deepcopy(kv.details)
            kv.details.update(d)
            if status is not None:
                kv.status = status
            if completion_time is not None:
                kv.completion_time = completion_time
        else:
            session.add(
                Job(
                    label=id,
                    details=d,
                    version=version,
                    status=status,
                    submit_time=submit_time,
                    completion_time=completion_time,
                    parent_label=parent_label,
                    job_name=job_name,
                )
            )
        session.commit()

    def exists(self, job_id: str) -> bool:
        """Check if there is a job in the database with this id

        :param job_id:
        :return:
        """
        return self._get_if(job_id) is not None

    def remove(self, job_id: str) -> bool:
        """Delete a job from the database

        :param job_id:
        :return: Did the removal succeed.
        """
        session = self.Session()
        children = session.query(Job).filter(Job.parent_label == job_id).all()
        for job in children:
            session.delete(job)

        parent = session.query(Job).filter(Job.label == job_id).first()
        session.delete(parent)

        session.commit()
        return True

    def parents_like(self, pattern: str) -> List[str]:
        """Get all the parent jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any parents that match this
        { <field>: { $regex: /pattern/<options> } }
        """
        session = self.Session()
        matches = [
            h.label
            for h in session.query(Job).filter(Job.parent_label == None).filter(Job.label.ilike(f'{pattern}%')).all()
        ]
        return matches

    def children_like(self, pattern: str) -> List[str]:
        """Get all jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any jobs that match this
        """
        session = self.Session()
        matches = [
            h.label
            for h in session.query(Job).filter(Job.parent_label != None).filter(Job.label.ilike(f'{pattern}%')).all()
        ]
        return matches

    def is_a_child(self, child: str) -> bool:
        """Is the resource a child of a job

        :param child: The name of the child
        :return: `True` if its a child
        """
        return self._get_if(child).first().parent_label is not None



import os
import re
from setuptools import setup, find_packages


def get_version(project_name):
    regex = re.compile(r"""^__version__ = ["'](\d+\.\d+\.\d+-?(?:a|b|rc|dev|alpha)?\.?(?:\d)*?)['"]$""")
    with open(f"{project_name}/version.py") as f:
        for line in f:
            m = regex.match(line.rstrip("\n"))
            if m is not None:
                return m.groups(1)[0]


class About(object):
    NAME = 'odin'
    AUTHOR = 'Interactions, LLC'
    VERSION = get_version(NAME)
    EMAIL = f"{NAME}@interactions.com"


setup(
    name=f"{About.NAME}-ml",
    version=About.VERSION,
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pyyaml>=5.1',
        'websockets',
        'kubernetes',
        'shortid',
        'GitPython',
        'pymongo',
        'SQLAlchemy',
        'psycopg2-binary',
        'ruamel.yaml',
        'cachetools',
        'prompt_toolkit >= 2.0.0',
        'requests-async',
        'requests',
        'pandas',
        'mead-baseline >= 2.0.1',
        'mead-xpctl-client',
    ],
    extras_require={'test': ['pytest']},
    entry_points={
        'console_scripts': [
            'odin-auth = odin.client.authenticate:main',
            'odin-user = odin.client.user:main',
            'odin-chores = odin.chores:main',
            'odin-cleanup = odin.client.cleanup:main',
            'odin-select-model = odin.model.selector:main',
            'odin-generate = odin.client.generate:main',
            'odin-logs = odin.client.logs:main',
            'odin-events = odin.client.events:main',
            'odin-run = odin.client.run:main',
            'odin-serve = odin.serve:main',
            'odin-status = odin.client.status:main',
            'odin-data = odin.client.data:main',
            'odin-show = odin.client.show:main',
            'odin-template = odin.template:main',
            'odin-ping = odin.client.ping:main',
            'odin-push = odin.client.push:main',
            'odin-create = odin.client.create:main',
            'odin-gpus = odin.client.gpus:main',
            'yaml2json = odin.utils.yaml:main',
            'json2yaml = odin.utils.yaml:json_main',
        ]
    },
)



"""A chore that summarizes and notifies the mead eval performance."""

import os
from typing import Dict
from string import Template
import requests
import pandas as pd
from odin.chores import register_chore


@register_chore('slack-eval-report-intent')
def slack_webhook_mead_eval_intent(parent_details: Dict, webhook: str, template: str, base_dir: str) -> None:
    """Substitute a template message and post to slack

    :param parent_details: The context to use to replace values in the template.
    :param webhook: The webhook key
    :param template: The message.
    :param base_dir: the base dir to read result files from
    """

    message: str = Template(template).substitute(parent_details)
    mismatched_jobs = []
    for job_id in parent_details['executed']:
        if 'mead-eval' in job_id:
            df = pd.read_csv(os.path.join(base_dir, f'{job_id}.tsv'), header=None, sep='\t')
            mismatched = df[df[1] != df[2]].copy()
            mismatched.insert(0, 'job_id', job_id)
            message = message + f'\n[[job_id]] {job_id} [[failures]]: {len(mismatched)}'
            os.remove(os.path.join(base_dir, f'{job_id}.tsv'))
            mismatched_jobs.append(mismatched)
    if mismatched_jobs:
        pd.concat(mismatched_jobs).to_csv(os.path.join(base_dir, 'results.csv'))
    requests.post(webhook, json={"text": message})



"""Create a Job on a remote odin server."""

import os
import json
import argparse
from getpass import getuser
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient
from odin.utils.auth import get_jwt_token


def create_job_http(url: str, jwt_token: str, name: str) -> None:
    """Request the server makes a new job.

    :param url: Base url of the remote odin server
    :param jwt_token: You JWT authentication token
    :param name: The name of the job you want to create
    """
    results = HttpClient(url, jwt_token).create_job(name)
    print(json.dumps(results))


def main():
    """Websocket client for pinging odin."""
    parser = argparse.ArgumentParser(description='Create a job')
    parser.add_argument('job', help="The name of the job you are creating")
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument('--token', help="File where JWT token can reside", default=os.path.expanduser("~/.odin.token"))
    parser.add_argument('--username', '-u', help="Username", default=getuser())
    parser.add_argument('--password', '-p', help="Password")
    parser.add_argument(
        '--scheme', choices={'http', 'https'}, default=ODIN_SCHEME, help='Connection protocol, use `http` for REST.'
    )
    args = parser.parse_args()
    url = f'{args.scheme}://{args.host}:{args.port}'

    jwt_token = get_jwt_token(url, args.token, args.username, args.password)
    try:
        create_job_http(url, jwt_token, args.job)
    except ValueError:
        if os.path.exists(args.token):
            os.remove(args.token)
            jwt_token = get_jwt_token(url, args.token, args.username, args.password)
            create_job_http(url, jwt_token, args.job)


if __name__ == "__main__":
    main()



"""Websocket client to check if odin is running."""

import argparse
import asyncio
from datetime import datetime
import json
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME


async def ping(uri: str, message: str) -> None:
    """Ping odin at uri and send message.

    :param uri: The location of the server
    :param message: The message you expect to see back
    :raises RuntimeError: If the server returns an error
    """
    async with websockets.connect(uri) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'PING', APIField.REQUEST: message}))
        resp = json.loads(await websocket.recv())
        if resp[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(resp)
            raise RuntimeError(resp)
        LOGGER.info(resp[APIField.RESPONSE])


def main():
    """Websocket client for pinging odin."""
    parser = argparse.ArgumentParser(description='Websocket-based health check')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws'},
        default=ODIN_SCHEME,
        help='Websocket connection protocol, use `wss` for remote connections and `ws` for localhost',
    )
    parser.add_argument('--message', default=datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
    args = parser.parse_args()
    ws = f'{args.scheme}://{args.host}:{args.port}'
    asyncio.get_event_loop().run_until_complete(ping(ws, args.message))


if __name__ == "__main__":
    main()



"""Tools to manipulate YAML
"""
import argparse
import json
import os
from typing import Dict, TextIO, Union
import yaml

from baseline.utils import str_file, str2bool


Path = str


@str_file(data='r', out='w')
def yaml_to_json(data: str, out: str, indent: int = 2) -> None:
    """Convert a YAML file into a JSON file

    :param data: The file to read from
    :param out: The file to write to
    :param indent: The indentation
    """
    json.dump(yaml.load(data, Loader=yaml.FullLoader), out, indent=indent)


@str_file(data='r', out='w')
def json_to_yaml(
    data: Union[Path, TextIO], out: Union[Path, TextIO], indent: int = 2, default_flow: bool = False
) -> None:
    """Convert a JSON file into a YAML file.

    :param data: The file to read from
    :param out: The file to write to
    :param indent: The indentation to use in the output
    :param default_flow: Should the yaml lists be inlined?
    """
    yaml.dump(json.load(data), out, indent=indent, default_flow_style=default_flow)


@str_file(file_path="w")
def write_yaml(content: Dict, file_path: Union[Path, TextIO]) -> None:
    """write data out to a yaml file.

    :param content: The data to be written.
    :param file_path: The file to write to.
    """
    yaml.dump(content, file_path, default_flow_style=False)


def main():
    """Convert a YAML file to JSON"""
    parser = argparse.ArgumentParser(description="Convert a yaml file to json.")
    parser.add_argument("file", help="The yaml file.")
    parser.add_argument("--indent", help="Number of spaces to indent in json.", default=2, type=int)
    parser.add_argument("--out", help="The name of the output file")
    args = parser.parse_args()

    if args.out is None:
        base, _ = os.path.splitext(args.file)
        args.out = f"{base}.json"

    yaml_to_json(args.file, args.out, indent=args.indent)


def json_main():
    """Convert a JSON file to YAML"""
    parser = argparse.ArgumentParser(description="Convert a json file to yaml.")
    parser.add_argument("file", help="The yaml file.")
    parser.add_argument("--indent", help="Number of spaces to indent in yaml.", default=2, type=int)
    parser.add_argument("--out", help="The name of the output file")
    parser.add_argument(
        "--default-flow", help="Should you use the yaml flow were lists are inline", default=False, type=str2bool
    )
    args = parser.parse_args()

    if args.out is None:
        base, _ = os.path.splitext(args.file)
        args.out = f"{base}.yml"

    json_to_yaml(args.file, args.out, indent=args.indent, default_flow=args.default_flow)



"""Defines (partial) resource handler for k8s Services"""

from typing import List
from kubernetes import client
from odin.k8s import ResourceHandler, json_to_selector, register_resource_handler


@register_resource_handler(aliases=['svc'])
class ServiceHandler(ResourceHandler):
    """(partial) resource handler for k8s Services
    """

    NAME = "Service"

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return ServiceHandler.NAME

    def get_api(self) -> object:
        """Get the service API

        :return: Service API
        :rtype object
        """
        return self.core_api

    def get_pods(self, name: str) -> List[client.models.v1_pod.V1Pod]:
        """Get the list of pods that are managed by a service.

        :param name: The name of the service you want the pods from.
        :type name: str
        :returns: The pods
        :rtype: List[client.models.v1_pod.V1Pod]
        """
        try:
            service = self.core_api.read_namespaced_service(name, self.namespace)
            selectors = service.spec.selector
            selectors = json_to_selector(selectors)
            return self.core_api.list_namespaced_pod(self.namespace, label_selector=selectors).items
        except client.rest.ApiException:
            return []
